@@ -22,6 +22,7 @@
 
 import __main__
 
+from functools import partial
 import sys
 import os
 from urllib.parse import quote as url_quote
@@ -30,13 +31,21 @@ import signal
 import subprocess
 import traceback
 import errno
+import fcntl
 import time
-import pipes
 import socket
 import shutil
+import shlex
+from typing import (List, Optional, Union, Tuple)
 
 import adtlog
 
+# When running installed, this is /usr/share/autopkgtest.
+# When running uninstalled, this is the source tree.
+# Either way, it has a lib subdirectory from which virtualization
+# backends can read extra files.
+PKGDATADIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
+
 progname = "<VirtSubproc>"
 devnull_read = open('/dev/null', 'rb')
 caller = __main__
@@ -57,15 +66,18 @@ class Quit(RuntimeError):
 
 
 class Timeout(RuntimeError):
-    pass
+
+    def __init__(self, timeout_secs, abort=False):
+        self.timeout_secs = timeout_secs
+        self.abort = abort
 
 
-def alarm_handler(*a):
-    raise Timeout()
+def alarm_handler(to, *a):
+    raise Timeout(to)
 
 
 def timeout_start(to):
-    signal.signal(signal.SIGALRM, alarm_handler)
+    signal.signal(signal.SIGALRM, partial(alarm_handler, to))
     signal.alarm(to)
 
 
@@ -194,6 +206,61 @@ def check_exec(argv, downp=False, outp=False, timeout=0, fail_on_stderr=True):
     return out
 
 
+def load_shell_script(path: str) -> str:
+    if not os.path.isabs(path):
+        path = os.path.join(PKGDATADIR, path)
+
+    with open(path) as reader:
+        return reader.read()
+
+
+def wait_booted(
+    exec_command: List[str],
+    timeout: int = 120,
+    short_command_timeout: int = 10,
+) -> None:
+    '''
+    Wait until a container or VM with an init system has sufficiently
+    booted to interact with it.
+
+    timeout: how long to wait for the testbed to complete boot
+
+    short_command_timeout: how long we will wait for a relatively quick command
+    to finish
+    '''
+    await_boot = load_shell_script('lib/in-testbed/await-boot.sh')
+
+    while timeout > 0:
+        timeout -= 1
+        time.sleep(1)
+        adtlog.debug('wait_booted: trying to execute trivial command on container')
+        try:
+            (rc, _, _) = execute_timeout(
+                None, short_command_timeout, exec_command + ['true'],
+            )
+        except Timeout:
+            # With some virt servers, commands can get stuck when executed
+            # while the testbed is booting, or at least this is what we observe.
+            # The wait_booted() retry loop has its own timeout, so we shouldn't
+            # stop on timeouts internal to the loop.
+            rc = 1
+        if rc == 0:
+            adtlog.debug("wait_booted: trivial command executed successfully")
+            break
+    else:
+        bomb('timed out waiting for testbed to start')
+
+    adtlog.debug('testbed is alive, waiting for boot to complete')
+    try:
+        check_exec(
+            exec_command + ['sh', '-ec', await_boot],
+            timeout=timeout,
+            fail_on_stderr=False,
+        )
+    except Timeout:
+        bomb('timed out waiting for testbed to complete boot')
+
+
 class timeout:
     def __init__(self, secs, exit_msg=None):
         '''Context manager that times out after given number of seconds.
@@ -229,9 +296,35 @@ def get_unix_socket(path):
     return s
 
 
-def expect(sock, search_bytes, timeout_sec, description=None, echo=False):
-    adtlog.debug('expect: "%s"' % (search_bytes or b'<none>').decode())
-    what = '"%s"' % (description or search_bytes or 'data')
+def matches_expectation(
+    haystack: bytes,
+    needle: Union[None, bytes, str, Tuple[Union[bytes, str]]],
+) -> bool:
+    if needle is None:
+        return True
+    elif isinstance(needle, tuple):
+        for n in needle:
+            if matches_expectation(haystack, n):
+                return True
+        else:
+            return False
+    elif isinstance(needle, bytes):
+        return needle in haystack
+    elif isinstance(needle, str):
+        return needle.encode('utf-8') in haystack
+    else:
+        raise TypeError(repr(needle))
+
+
+def expect(
+    sock: socket.socket,
+    search: Union[None, bytes, str, Tuple[Union[bytes, str]]],
+    timeout_sec: int,
+    description: Optional[str] = None,
+    echo: bool = False,
+):
+    adtlog.debug('expect: %r' % (search,))
+    what = repr(description or search or 'data')
     out = b''
     with timeout(timeout_sec,
                  description and ('timed out waiting for %s' % what) or None):
@@ -243,7 +336,8 @@ def expect(sock, search_bytes, timeout_sec, description=None, echo=False):
             if echo:
                 sys.stderr.buffer.write(block)
             out += block
-            if search_bytes is None or search_bytes in out:
+
+            if matches_expectation(out, search):
                 adtlog.debug('expect: found "%s"' % what)
                 break
 
@@ -274,13 +368,13 @@ def downtmp_mktemp(path):
     fail if they get moved around.
     '''
     if path:
-        check_exec(['mkdir', '--mode=1777', '--parents', downtmp_open],
+        check_exec(['/bin/mkdir', '--mode=1777', '--parents', downtmp_open],
                    downp=True)
         return path
     else:
-        d = check_exec(['mktemp', '--directory', '--tmpdir', 'autopkgtest.XXXXXX'],
+        d = check_exec(['/bin/mktemp', '--directory', '--tmpdir', 'autopkgtest.XXXXXX'],
                        downp=True, outp=True)
-        check_exec(['chmod', '1777', d], downp=True)
+        check_exec(['/bin/chmod', '1777', d], downp=True)
         return d
 
 
@@ -309,9 +403,31 @@ def cmd_revert(c, ce):
     return [downtmp]
 
 
+def reboot_testbed():
+    (systemd_check, out, err) = execute_timeout(
+        None, 10, auxverb + ['test', '-d', '/run/systemd/system'],
+        stdout=subprocess.PIPE, stderr=subprocess.PIPE
+    )
+    if systemd_check == 0:
+        reboot_cmd = ['systemd-run', '--no-block', '--quiet', 'sh', '-c', 'sleep 3; reboot']
+        adtlog.debug('rebooting with command: %s' % reboot_cmd)
+        status, out, err = execute_timeout(None, 30, auxverb + reboot_cmd)
+
+        if status == 0:
+            return
+        else:
+            adtlog.debug(
+                'Cannot reboot with %s, trying something else' % reboot_cmd)
+
+    reboot_cmd = ['sh', '-c', '(sleep 3; reboot) >/dev/null 2>&1 &']
+    adtlog.debug('rebooting with command: %s' % reboot_cmd)
+    status, out, err = execute_timeout(None, 30, auxverb + reboot_cmd)
+
+
 def cmd_reboot(c, ce):
     global downtmp
     cmdnumargs(c, ce, 0, 1)
+    wait_reboot_args = {}
     if not downtmp:
         bomb("`reboot' when not open")
     if 'reboot' not in caller.hook_capabilities():
@@ -320,16 +436,16 @@ def cmd_reboot(c, ce):
     # save current downtmp; try a few locations, as /var/cache might be r/o
     # (argh Ubuntu touch)
     directories = '/var/cache /home'
-    check_exec(['sh', '-ec', 'for d in %s; do if [ -w $d ]; then '
-                '  tar --warning=none --create --absolute-names '
+    check_exec(['/bin/sh', '-ec', 'for d in %s; do if [ -w $d ]; then '
+                '  /bin/tar --warning=none --create --absolute-names '
                 '''    -f $d/autopkgtest-tmpdir.tar '%s'; '''
-                '  rm -f /run/autopkgtest-reboot-prepare-mark; '
-                '  exit 0; fi; done; exit 1''' % (directories, downtmp)],
+                '  /bin/rm -f /run/autopkgtest-reboot-prepare-mark; '
+                '  exit 0; fi; done; exit 1' % (directories, downtmp)],
                downp=True, timeout=copy_timeout)
     adtlog.debug('cmd_reboot: saved current downtmp, rebooting')
 
     try:
-        caller.hook_prepare_reboot()
+        wait_reboot_args = caller.hook_prepare_reboot() or {}
     except AttributeError:
         pass
 
@@ -337,16 +453,16 @@ def cmd_reboot(c, ce):
     if len(c) > 1 and c[1] == 'prepare-only':
         adtlog.info('state saved, waiting for testbed to reboot...')
     else:
-        execute_timeout(None, 30, auxverb +
-                        ['sh', '-c', '(sleep 3; reboot) >/dev/null 2>&1 &'])
-    caller.hook_wait_reboot()
+        reboot_testbed()
+
+    caller.hook_wait_reboot(**wait_reboot_args)
 
     # restore downtmp
-    check_exec(['sh', '-ec', 'for d in %s; do '
+    check_exec(['/bin/sh', '-ec', 'for d in %s; do '
                 'if [ -e $d/autopkgtest-tmpdir.tar ]; then '
-                ' tar --warning=none --extract --absolute-names '
+                ' /bin/tar --warning=none --extract --absolute-names '
                 '     -f $d/autopkgtest-tmpdir.tar;'
-                ' rm $d/autopkgtest-tmpdir.tar; exit 0; '
+                ' /bin/rm $d/autopkgtest-tmpdir.tar; exit 0; '
                 'fi; done; exit 1' % directories],
                downp=True, timeout=copy_timeout)
     adtlog.debug('cmd_reboot: restored downtmp after reboot')
@@ -392,7 +508,7 @@ def copyup_shareddir(tb, host, is_dir, downtmp_host):
             tb_tmp = os.path.join(downtmp, os.path.basename(host))
             adtlog.debug('copyup_shareddir: tb path %s is not already in '
                          'downtmp, copying to %s' % (tb, tb_tmp))
-            check_exec(['cp', '-r', '--preserve=timestamps,links', tb, tb_tmp],
+            check_exec(['/bin/cp', '-r', '--preserve=timestamps,links', tb, tb_tmp],
                        downp=True)
             # translate into host path
             tb = os.path.join(downtmp_host, os.path.basename(host))
@@ -409,7 +525,7 @@ def copyup_shareddir(tb, host, is_dir, downtmp_host):
 
         if tb_tmp:
             adtlog.debug('copyup_shareddir: rm intermediate copy: %s' % tb)
-            check_exec(['rm', '-rf', tb_tmp], downp=True)
+            check_exec(['/bin/rm', '-rf', tb_tmp], downp=True)
     finally:
         timeout_stop()
 
@@ -455,8 +571,8 @@ def copydown_shareddir(host, tb, is_dir, downtmp_host):
         if host == tb:
             host_tmp = None
         else:
-            check_exec(['rm', '-rf', tb], downp=True)
-            check_exec(['cp', '-r', '--preserve=timestamps,links', host, tb],
+            check_exec(['/bin/rm', '-rf', tb], downp=True)
+            check_exec(['/bin/cp', '-r', '--preserve=timestamps,links', host, tb],
                        downp=True)
         if host_tmp:
             (is_dir and shutil.rmtree or os.unlink)(host_tmp)
@@ -508,7 +624,7 @@ def copyupdown_internal(wh, sd, upp):
 
     deststdout = devnull_read
     srcstdin = devnull_read
-    remfileq = pipes.quote(sd[iremote])
+    remfileq = shlex.quote(sd[iremote])
     if not dirsp:
         rune = 'cat %s%s' % ('><'[upp], remfileq)
         if upp:
@@ -566,8 +682,9 @@ def copyupdown_internal(wh, sd, upp):
             status = subprocs[sdn].wait()
             if not (status == 0 or (sdn == 0 and status == -13)):
                 timeout_stop()
-                bomb("%s %s failed, status %d" %
-                     (wh, ['source', 'destination'][sdn], status))
+                adtlog.info("%s %s failed, status %d" %
+                            (wh, ['source', 'destination'][sdn], status))
+                raise FailedCmd(['copy-failed'])
         timeout_stop()
     except Timeout:
         for sdn in [1, 0]:
@@ -614,22 +731,7 @@ def cmd_shell(c, ce):
 
 def command():
     sys.stdout.flush()
-    while True:
-        try:
-            ce = sys.stdin.readline().strip()
-            # FIXME: This usually means EOF (as checked below), but with Python
-            # 3 we often get empty strings here even though this is supposed to
-            # block for new input.
-            if ce == '':
-                time.sleep(0.1)
-                continue
-            break
-        except IOError as e:
-            if e.errno == errno.EAGAIN:
-                time.sleep(0.1)
-                continue
-            else:
-                raise
+    ce = sys.stdin.readline()
     if not ce:
         bomb('end of file - caller quit?')
     ce = ce.rstrip().split()
@@ -652,8 +754,8 @@ def command():
     print(' '.join(r))
 
 
-signal_list = [	signal.SIGHUP, signal.SIGTERM,
-                signal.SIGINT, signal.SIGPIPE]
+signal_list = [signal.SIGHUP, signal.SIGTERM,
+               signal.SIGINT, signal.SIGPIPE]
 
 
 def sethandlers(f):
@@ -710,6 +812,20 @@ def cmd_auxverb_debug_fail(c, ce):
 
 def mainloop():
     global in_mainloop
+
+    if (fcntl.fcntl(0, fcntl.F_GETFL) & os.O_NONBLOCK) != 0:
+        # Some versions of autopkgtest-virt-* would try to cope
+        # by sleeping 0.1 when they got EOF/EAGAIN.
+        # But that causes us to hang, looping, when our caller goes away.
+        #
+        # Even though this is an egregious violation of Unix calling
+        # conventions, and would be entirely the caller's fault,
+        # detecting this situation and crashing is sensible.
+        # Otherwise we would produce misleading error messages -
+        # stdin.readline() returns '', not EAGAIN, which looks just like EOF.
+        # (And in theory, we might read, and act on, partial commands!)
+        bomb('stdin to autopkgtest virt server is nonblocking!')
+
     in_mainloop = True
 
     try:
