@@ -92,6 +92,7 @@ class Testbed:
         *,
         vserver_argv: List[str],
         output_dir: str,
+        test_arch: Optional[str] = None,
         user: str,
         setup_commands: List[str],
         setup_commands_boot: List[str],
@@ -113,6 +114,7 @@ class Testbed:
         self._need_reset_apt = False
         self.stop_sent = False
         self.dpkg_arch = None
+        self.test_arch = test_arch
         self.exec_cmd = None
         self.output_dir = output_dir
         self.shared_downtmp = None  # testbed's downtmp on the host, if supported
@@ -356,6 +358,12 @@ class Testbed:
         self.dpkg_arch = self.check_exec(['dpkg', '--print-architecture'], True).strip()
         adtlog.info('testbed dpkg architecture: ' + self.dpkg_arch)
 
+        # the test architecture defaults to the dpkg architecture
+        if not self.test_arch:
+            self.test_arch = self.dpkg_arch
+
+        self.test_arch_is_foreign = (self.test_arch != self.dpkg_arch)
+
         out = self.check_exec(['dpkg-query', '-W', '-f', '${Version}', 'apt'], True).strip()
         adtlog.info('testbed apt version: ' + out)
         try:
@@ -363,6 +371,17 @@ class Testbed:
         except ValueError:
             self.bomb("can't determine the apt version on the testbed")
 
+        # Foreign arch testing requires `apt-get satisfy`.
+        if self.test_arch_is_foreign and self.apt_version < Version("1.9.0"):
+            self.bomb(
+                "foreign arch testing requires apt >= 1.9.0 on the testbed",
+                adtlog.AutopkgtestError,
+            )
+
+        # Add foreign test architecture
+        if self.test_arch_is_foreign:
+            self.check_exec(['dpkg', '--add-architecture', self.test_arch])
+
         # do we have eatmydata?
         (code, out, err) = self.execute(
             ['sh', '-ec', 'command -v eatmydata'],
@@ -422,14 +441,6 @@ class Testbed:
         if self.default_release:
             self._set_default_release()
 
-        for source in self.add_apt_sources:
-            if self.SOURCELIST_PATTERN.match(source):
-                self._add_apt_source(source, filename='autopkgtest-add-apt-sources.list')
-            elif self.PPA_PATTERN.match(source):
-                self._add_apt_ppa(source)
-            else:
-                self.bomb(f"invalid apt source: {source}", adtlog.AutopkgtestError)
-
         pin_packages = self.pin_packages.copy()
         add_apt_releases = self.add_apt_releases.copy()
         for pocket in self.add_apt_pockets:
@@ -457,6 +468,26 @@ class Testbed:
             # Set baseline "general form" pin to neutralize NotAutomatic: yes.
             self._set_baseline_pin_priority_for_release(release)
 
+        # To be done after adding add_apt_releases, so that adding releases/pocket
+        # is only done based on sources the image came configured with (LP: #2091393).
+        for source in self.add_apt_sources:
+            if self.SOURCELIST_PATTERN.match(source):
+                self._add_apt_source(source, filename='autopkgtest-add-apt-sources.list')
+            elif self.PPA_PATTERN.match(source):
+                self._add_apt_ppa(source)
+            elif (
+                os.path.isfile(source) and
+                os.access(source, os.R_OK) and
+                (source.endswith('.list') or source.endswith('.sources'))
+            ):
+                with open(source, 'r') as s:
+                    self._add_apt_source(
+                        s.read(),
+                        filename=os.path.basename(source),
+                    )
+            else:
+                self.bomb(f"invalid apt source: {source}", adtlog.AutopkgtestError)
+
         # apt-get update, to be done after adding sources, releases, pockets,
         # but before setting up pinning for apt (<< 1.9.11) compatibility, as
         # it requires src:package expansion via apt-cache dumpavail.
@@ -545,8 +576,16 @@ class Testbed:
             self._opened(pl)
         self.modified = False
 
-    def install_deps(self, deps_new, shell_on_failure=False, synth_deps=[]):
-        '''Install dependencies into testbed'''
+    def install_deps(
+        self,
+        deps_new: List[List[str]],
+        shell_on_failure: bool = False,
+        package_under_test_deps: Optional[List[List[str]]] = None,
+    ) -> None:
+        """
+        Install test dependencies into testbed.
+        """
+
         adtlog.debug('install_deps: deps_new=%s' % deps_new)
 
         self.deps_installed = deps_new
@@ -554,10 +593,10 @@ class Testbed:
             return
 
         self.satisfy_dependencies_string(
-            ', '.join(deps_new),
+            ', '.join([' | '.join(altgroup) for altgroup in deps_new]),
             'install-deps',
             shell_on_failure=shell_on_failure,
-            synth_deps=synth_deps
+            package_under_test_deps=package_under_test_deps,
         )
 
     def _provide_sudo(self) -> str:
@@ -913,9 +952,12 @@ class Testbed:
             else:
                 raise adtlog.AptPermanentError
 
-    def _install_apt_synth_deps(self, synth_deps: List[List[str]]) -> None:
+    def _install_apt_package_under_test_deps(
+        self,
+        package_under_test_deps: List[List[str]]
+    ) -> None:
         need_explicit_install = []
-        for dep in synth_deps:
+        for dep in package_under_test_deps:
             if len(dep) == 1:
                 # simple test dependency (no alternatives)
                 pkg = dep[0]
@@ -981,7 +1023,7 @@ class Testbed:
         self,
         deps: str,
         shell_on_failure: bool = False,
-        synth_deps: Optional[List[List[str]]] = None,
+        package_under_test_deps: Optional[List[List[str]]] = None,
     ) -> None:
         '''Install dependencies with apt-get into testbed
 
@@ -1039,11 +1081,12 @@ class Testbed:
                     self._run_apt_command(action='install', extra_opts=['--fix-broken'])
                 elif deps:
                     self._run_apt_command(action='satisfy', what=[deps])
-                # if '@' was provided as a test dependency (as seen in synth_deps),
+                # For dependencies that are part of the source package under
+                # test (for example if '@' was provided as a test dependency),
                 # we explicitly install the corresponding binary packages in case
                 # the dependencies on those were satisfied by versioned Provides
-                if synth_deps:
-                    self._install_apt_synth_deps(synth_deps)
+                if package_under_test_deps:
+                    self._install_apt_package_under_test_deps(package_under_test_deps)
 
             except adtlog.AptDownloadError:
                 self.bomb('apt failed to download packages')
@@ -1165,7 +1208,13 @@ class Testbed:
                 self.run_shell()
             self.bomb("Failure installing build dependencies on testbed")
 
-    def satisfy_dependencies_string(self, deps, what, shell_on_failure=False, synth_deps=[]):
+    def satisfy_dependencies_string(
+        self,
+        deps: str,
+        what: str,
+        shell_on_failure: bool = False,
+        package_under_test_deps: Optional[List[List[str]]] = None,
+    ) -> None:
         '''Install dependencies from a string into the testbed'''
 
         adtlog.debug('%s: satisfying %s' % (what, deps))
@@ -1177,7 +1226,7 @@ class Testbed:
         adtlog.debug('can use apt-get on testbed: %s' % can_apt_get)
 
         if can_apt_get:
-            self.install_apt(deps, shell_on_failure, synth_deps)
+            self.install_apt(deps, shell_on_failure, package_under_test_deps)
         else:
             # This whole `else` branch is meant to check if we can get
             # away without being root on the testbed because all the
@@ -1284,6 +1333,8 @@ class Testbed:
         argv.extend([
             '--artifacts={}'.format(test_artifacts),
             '--chdir={}'.format(tree.tb),
+            '--env=AUTOPKGTEST_TESTBED_ARCH={}'.format(self.dpkg_arch),
+            '--env=AUTOPKGTEST_TEST_ARCH={}'.format(self.test_arch),
             '--env=DEB_BUILD_OPTIONS=parallel={}'.format(
                 build_parallel or self.nproc
             ),
@@ -1310,6 +1361,20 @@ class Testbed:
             '--tmp={}'.format(autopkgtest_tmp),
         ])
 
+        if self.test_arch_is_foreign:
+            rc, dpkg_arch_vars, _ = self.execute(
+                ['dpkg-architecture', '-a' + self.test_arch],
+                stdout=subprocess.PIPE,
+                stderr=subprocess.DEVNULL,
+            )
+            testbed_has_dpkg_architecture = (rc == 0)
+            if testbed_has_dpkg_architecture:
+                for var in dpkg_arch_vars.splitlines():
+                    # set up environment for cross-architecture building.
+                    if var.startswith('DEB_HOST_') or var.startswith('DEB_BUILD_'):
+                        argv.append(f'--env={var}')
+                adtlog.info('test environment configured for cross building')
+
         if 'needs-root' in test.restrictions and self.user is not None:
             argv.append('--env=AUTOPKGTEST_NORMAL_USER={}'.format(self.user))
             argv.append('--env=ADT_NORMAL_USER={}'.format(self.user))
@@ -1527,11 +1592,14 @@ class Testbed:
                 """
             )
         )
+
+        pkglist = pkglist.split(',')
+
         if self.apt_version < Version('2.0'):
             # sort pkglist into source and binary packages
             binpkgs = []
             srcpkgs = []
-            for i in pkglist.split(','):
+            for i in pkglist:
                 i = i.strip()
                 if i.startswith('src:'):
                     srcpkgs.append(i[4:])
@@ -1552,7 +1620,13 @@ class Testbed:
                     if src in srcpkgs and pkg not in binpkgs:
                         binpkgs.append(pkg)
 
-            pkglist = ",".join(binpkgs)
+            pkglist = binpkgs
+
+        # We want pins to be valid on any architecture, so that they work
+        # consistently across architectures in multi-arch systems.
+        for i, pkg in enumerate(pkglist):
+            if ':' not in (pkg[4:] if pkg.startswith('src:') else pkg):
+                pkglist[i] += ':any'
 
         # Pin the release with priority 100, which is the same as if it had
         # "NotAutomatic: yes", "ButAutomaticUpgrades: yes", which is what we
@@ -1566,9 +1640,10 @@ class Testbed:
         # prefer given packages from the specified release
         if pkglist:
             pref_content += pref_template.substitute(
-                pkgs=pkglist.replace(',', ' '),
+                pkgs=' '.join(pkglist),
                 release=release,
-                prio='995')
+                prio='995',
+            )
 
         pref_file = f"autopkgtest-{release}.pref"
         self._add_apt_preference(pref_content, pref_file)
@@ -1843,7 +1918,7 @@ class Testbed:
             ppa_repo_uri = urllib.parse.urlunparse(parsed_uri)
 
         # construct source list for ppa
-        if self.apt_version < Version("1.1~exp9"):
+        if self.apt_version < Version("2.3.10"):
             sourcelist = f"autopkgtest-ppa-{ppa['owner']}-{ppa['name']}.list"
 
             # add key to APT keyring
@@ -1859,7 +1934,8 @@ class Testbed:
                     "add_apt_ppa",
                     signed_by,
                     ppa_keyring,
-                ]
+                ],
+                xenv=["APT_KEY_DONT_WARN_ON_DANGEROUS_USAGE=1"]
             )
 
             # construct one-line-style source list
@@ -1898,6 +1974,128 @@ class Testbed:
         bad_chars = re.escape(r'~+=,()[]{}!@#$%&\'":;|\/')
         return re.sub(f'[{bad_chars}]', '_', filename)
 
+    def get_policy_for_package(self, package: str) -> Optional[str]:
+        """Return apt policy for package"""
+
+        (rc, stdout, _) = self.execute(
+            [
+                'apt-cache',
+                'policy',
+                # . and + are valid package name characters but are
+                # interpreted by $(apt-cache policy) as regex
+                '^' + package.replace('.', '\\.').replace('+', '\\+') + '$',
+            ],
+            stdout=subprocess.PIPE,
+        )
+
+        if rc != 0 or not stdout:
+            return None
+
+        return stdout
+
+    def get_candidate(self, package: str) -> Tuple[Optional[Version], Optional[int]]:
+        """Return candidate version and priority for package"""
+
+        policy = self.get_policy_for_package(package)
+        if not policy:
+            return None, None
+
+        # Get the candidate version of this binary
+        version = None
+        for line in policy.splitlines():
+            if not line.strip().startswith('Candidate:'):
+                continue
+            if "none" in line.lower():
+                break
+            version = line.strip().replace('Candidate: ', '')
+            break
+
+        if not version:
+            return None, None
+
+        priority = self._get_priority_from_policy(policy, version)
+
+        return Version(version), priority
+
+    def _get_priority_from_policy(self, policy: str, version: str) -> int:
+        """Extract the priority for a package version from its policy.
+
+        :param policy: `apt-cache policy` output for a single package
+        :type policy: str
+        :param version: the version for which to extract the prioriy
+        :type version: str
+        :return: priority for the given version
+        :rtype: int
+        """
+
+        # The apt-policy output for a package looks like this:
+        #
+        # apt:
+        #   Installed: 2.7.14build2
+        #   Candidate: 2.7.14build2
+        #   Version table:
+        #      2.9.2 100
+        #         100 http://archive.ubuntu.com/ubuntu oracular-proposed/main amd64 Packages
+        #      2.9.1 50
+        #         100 http://archive.ubuntu.com/ubuntu oracular-baz/main amd64 Packages
+        #  *** 2.7.14build2 500 (phased 20%)
+        #         500 http://archive.ubuntu.com/ubuntu oracular/main amd64 Packages
+        #         100 /path/to/dpkg/status
+        #      0.0.1 -30000
+        #        -30000 http://archive.ubuntu.com/ubuntu oracular-foo/main amd64 Packages
+        #      0.0.2 1
+        #           1 http://archive.ubuntu.com/ubuntu oracular-bar/main amd64 Packages
+        #
+        # We'll proceed in stages:
+        # 1. extract the "Version table"
+        # 2. from that, extract the entry for the desired version
+        # 3. from that, extract the version priority from the "version line" (first line)
+        # If priority != 0 (always true with apt >= 1.1~exp9) we are done, otherwise:
+        # 4. extract the max priority among "file" priorities.
+        #
+        # Notes:
+        # - The '***' indicates the currently installed version, and may or may
+        #   not be present.
+        # - For apt (<< 1.1~exp9) compatibility, we can't just extract the priority
+        #   from the line containing the package version, as apt (<< 1.1~exp9) just
+        #   prints 0 there if there is no "version pin" pin set.
+
+        # 1. Extract the version table. Note that apt prints "file priorities"
+        # with this format string: "       %4i %s\n", see private-show.cc.
+        needle = re.search(
+            r"(?<=^  Version table:\n)(^( {5}| \*{3} | {7,})[^ ].*\n)+",
+            policy,
+            flags=re.M
+        )
+        assert needle is not None
+        version_table = needle.group(0)
+
+        # 2. Extract the version table entry for the specified version. Refer to
+        # the apt format string mentioned above to see why we look for "^ {7,}".
+        needle = re.search(
+            r"^( {5}| \*{3} )" + re.escape(version) + " .*\n(^ {7,}[^ ].*\n)+",
+            version_table,
+            flags=re.M
+        )
+        assert needle is not None, \
+            "version_table doesn't contain version '%s':\n%s\npolicy:\n%s" % (version, version_table, policy)
+        version_table_for_candidate = needle.group(0)
+
+        # 3. Extract the "version line" and from that the priority.
+        version_line = version_table_for_candidate.splitlines()[0]
+        priority = int(version_line[5:].split()[1])
+
+        # This will always be true with apt >= 1.1~exp9.
+        if priority != 0:
+            return priority
+
+        # 5. Fallback to parsing "file priorities".
+        file_priorities = version_table_for_candidate.splitlines()[1:]
+        priorities = [int(fp.split(maxsplit=1)[0]) for fp in file_priorities]
+        priority = max(priorities)
+        assert priority != 0, "invalid priority"
+        return priority
+
 
 class TestbedPath:
     '''Represent a file/dir with a host and a testbed path'''
