@@ -160,7 +160,7 @@ testsrc can be one of a:
 You can specify local *.deb packages to test.'''
 
     epilog = '''The -- argument separates the autopkgtest actions and options
-from the virt-server which provides the testbed. See e. g. man autopkgtest-schroot
+from the virt-server which provides the testbed. See e. g. man autopkgtest-virt-qemu
 for details.'''
 
     parser = argparse.ArgumentParser(
@@ -182,6 +182,9 @@ for details.'''
     g_test.add_argument('-B', '--no-built-binaries', dest='built_binaries',
                         action='store_false', default=True,
                         help='do not build/use binaries from .dsc, git source, or unbuilt tree')
+    g_test.add_argument('-a', '--test-architecture', metavar='ARCH',
+                        help='run tests for (and when asked, build binaries '
+                        'for) ARCH instead of the testbed\'s architecture.')
     g_test.add_argument('packages', nargs='*',
                         help='testsrc source package and testbinary packages as above')
 
