@@ -169,6 +169,30 @@ def get_cpuflag() -> Sequence[str]:
     return []
 
 
+def get_host_time_zone() -> str:
+    '''Return the host system's time zone, or an empty string.'''
+    try:
+        target = os.readlink('/etc/localtime')
+
+        if '/zoneinfo/' in target:
+            return target.split('/zoneinfo/')[-1]
+    except OSError:
+        pass
+
+    try:
+        with open('/etc/timezone', 'r') as f:
+            for line in f:
+                if line.startswith('#'):
+                    continue
+                line = line.strip()
+                if line:
+                    return line
+    except OSError:
+        pass
+
+    return ''
+
+
 class QemuImage:
     def __init__(
         self,
