@@ -51,6 +51,7 @@ known_restrictions = frozenset({
     'needs-root',
     'needs-sudo',
     'rw-build-tree',
+    'skip-foreign-architecture',
     'skip-not-installable',
     'skippable',
     'superficial',
@@ -77,10 +78,13 @@ RESTRICTIONS_REQUIRE_CAPS: Dict[str, FrozenSet[Union[str, FrozenSet[str]]]] = {
 # Keys are restrictions
 # Values are sets of dependency strings
 RESTRICTIONS_IMPLY_DEPS = {
-    'needs-sudo': frozenset({'sudo'}),
+    'needs-sudo': frozenset({'passwd', 'sudo'}),
 }
 
 
+PKGDATADIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
+
+
 class Unsupported(Exception):
     '''Test cannot be run in the testbed'''
 
@@ -122,8 +126,8 @@ class Test:
         command: Optional[str],
         restrictions: Iterable[str],
         features: Iterable[str],
-        depends: List[str],
-        synth_depends: List[List[str]],
+        depends: List[List[str]],
+        package_under_test_depends: List[List[str]],
     ) -> None:
         '''Create new test description
 
@@ -135,7 +139,8 @@ class Test:
         @command: shell command for the test code
         @restrictions, @features: sets of strings, as in README.package-tests
         @depends: string list of test dependencies (packages)
-        @synth_depends: string list of synthesized test dependencies (packages)
+        @package_under_test_depends: string list with the subset of @depends
+        that come from the source package we are testing
         '''
         if '/' in name:
             raise Unsupported(name, 'test name may not contain / character')
@@ -149,12 +154,12 @@ class Test:
         self.restrictions = set(restrictions)
         self.features = set(features)
         self.depends = list(depends)
-        self.synth_depends = synth_depends
+        self.package_under_test_depends = package_under_test_depends
 
         for r in self.restrictions:
             for d in RESTRICTIONS_IMPLY_DEPS.get(r, frozenset()):
                 if d not in self.depends:
-                    self.depends.append(d)
+                    self.depends.append([d])
 
         # None while test hasn't run yet; True: pass, False: fail
         self.result = None
@@ -289,9 +294,38 @@ def _debian_check_unknown_fields(name, record):
         raise Unsupported(name, 'unknown field %s' % unknown_keys.pop())
 
 
+def _architecture_is_concerned(
+    architecture: str,
+    architecture_restrictions: List[str],
+) -> bool:
+    """Wrapper around python-debian's architecture_is_concerned() with perl
+    fallback in case python-debian is too old."""
+
+    # TODO: Add Depends: python3-debian (>= 0.1.48) to bin:autopkgtest and drop this.
+
+    if hasattr(debian.debian_support, "DpkgArchTable") and hasattr(
+        debian.debian_support.DpkgArchTable, "architecture_is_concerned"
+    ):
+        dpkg_arch_table = debian.debian_support.DpkgArchTable.load_arch_table()
+        return dpkg_arch_table.architecture_is_concerned(
+            architecture,
+            architecture_restrictions,
+        )
+    else:
+        res = subprocess.run(
+            [
+                os.path.join(PKGDATADIR, "lib", "arch-is-concerned.pl"),
+                architecture,
+                " ".join(architecture_restrictions),
+            ]
+        )
+        return res.returncode == 0
+
+
 def _debian_packages_from_source(
     srcdir: str,
     test_arch: str,
+    test_arch_is_foreign: bool,
 ) -> Tuple[List[str], List[str]]:
     """Parse source tree in srcdir and return tuple with: list of all binary
     deb packages built from that source, and the same list but restricted to
@@ -314,20 +348,11 @@ def _debian_packages_from_source(
         if arch == 'all':
             packages_for_test_arch.append(binarydeb)
         else:
-            if '-' not in test_arch:
-                # normalize test_arch to os-arch format, default to linux, Policy §11.1.
-                test_arch = 'linux-' + test_arch
-            (test_arch_os, test_arch_cpu) = test_arch.split('-')
-
             arch_list = arch.split()
-            arch_list = [a if '-' in a else 'any-' + a for a in arch_list]
-            for archspec in arch_list:
-                (arch_os, arch_cpu) = archspec.split('-')
-                if (arch_os == 'any' or arch_os == test_arch_os) and (
-                    arch_cpu == 'any' or arch_cpu == test_arch_cpu
-                ):
-                    packages_for_test_arch.append(binarydeb)
-                    break
+            if _architecture_is_concerned(test_arch, arch_list):
+                if test_arch_is_foreign:
+                    binarydeb += ':' + test_arch
+                packages_for_test_arch.append(binarydeb)
 
     return (packages, packages_for_test_arch)
 
@@ -362,15 +387,16 @@ def _debian_recommends_from_source(srcdir: str) -> List[str]:
     return deps
 
 
-def _debian_build_deps_from_source(srcdir: str) -> List[str]:
+def _debian_build_deps_from_source(srcdir: str) -> List[List[str]]:
     deps = []
     for st in parse_rfc822(os.path.join(srcdir, 'debian/control')):
         for key in ('Build-depends', 'Build-depends-indep', 'Build-depends-arch'):
             if key in st:
-                deps += [d.strip() for d in st[key].strip(',').split(',')]
+                altgr = [g.strip() for g in st[key].strip(',').split(',')]
+                deps += [[d.strip() for d in g.split('|')] for g in altgr]
 
     # @builddeps@ should always imply build-essential
-    deps.append('build-essential')
+    deps.append(['build-essential'])
     return deps
 
 
@@ -407,19 +433,25 @@ def _expand_test_depends(
     dep_str: str,
     srcdir: str,
     test_arch: str,
-) -> Tuple[List[str], List[List[str]]]:
+    test_arch_is_foreign: bool,
+) -> Tuple[List[List[str]], List[List[str]]]:
     '''Split test dependencies (comma separated), validate their syntax, and
     expand @, @builddeps@ and @recommends@. Returns list of dependencies
-    and list of synthesised dependencies.
+    and the subset of those dependencies that come from the source
+    package under test.
+
+    If test_arch_is_foreign is True, then :test_arch architecture qualifiers
+    are added to packages that are not Architecture: all.
 
     This may raise an InvalidControl exception if there are invalid
     dependencies.
     '''
     deps = []
-    synthdeps = []
+    package_under_test_deps = []
     (my_packages, my_packages_for_test_arch) = _debian_packages_from_source(
         srcdir,
         test_arch,
+        test_arch_is_foreign,
     )
 
     for alt_group_str in dep_str.split(','):
@@ -430,31 +462,50 @@ def _expand_test_depends(
         adtlog.debug('processing dependency %s' % alt_group_str)
         if alt_group_str == '@':
             for d in my_packages_for_test_arch:
-                adtlog.debug('synthesised dependency %s' % d)
-                deps.append(d)
-                synthdeps.append([d])
+                adtlog.debug('synthesised dependency %s from @' % d)
+                deps.append([d])
+                package_under_test_deps.append([d])
         elif alt_group_str == '@builddeps@':
-            for d in _debian_build_deps_from_source(srcdir):
-                adtlog.debug('synthesised dependency %s' % d)
-                deps.append(d)
+            for daltgr in _debian_build_deps_from_source(srcdir):
+                adtlog.debug('synthesised dependency %s from @builddeps@' % " | ".join(daltgr))
+                deps.append(daltgr)
         elif alt_group_str == '@recommends@':
             for d in _debian_recommends_from_source(srcdir):
-                adtlog.debug('synthesised dependency %s' % d)
-                deps.append(d)
+                adtlog.debug('synthesised dependency %s from @recommends@' % d)
+                deps.append([d])
         else:
             alt_group_list = []
-            synthdep_alternatives = []
+            package_under_test_alternatives = []
             for dep in [d.strip() for d in alt_group_str.split('|')]:
                 (pkg, version) = _debian_check_dep(testname, dep)
+                if test_arch_is_foreign and f'{pkg}:{test_arch}' in my_packages_for_test_arch:
+                    dep = dep.replace(pkg, f'{pkg}:{test_arch}', 1)
+                    pkg += f':{test_arch}'
                 alt_group_list.append(dep)
                 if pkg in my_packages:
-                    synthdep_alternatives.append(pkg)
-            if len(alt_group_list) == len(synthdep_alternatives):
-                adtlog.debug('marked alternatives %s as a synthesised dependency' % synthdep_alternatives)
-                synthdeps.append(synthdep_alternatives)
-            deps.append(' | '.join(alt_group_list))
+                    package_under_test_alternatives.append(pkg)
+            if len(alt_group_list) == len(package_under_test_alternatives):
+                # We depend on x|y|z, and all of x, y and z are binary
+                # packages built by the package under test
+                adtlog.debug('marked alternatives %s as part of the package under test' % package_under_test_alternatives)
+                package_under_test_deps.append(package_under_test_alternatives)
+            deps.append(alt_group_list)
 
-    return (deps, synthdeps)
+    if test_arch_is_foreign and ["build-essential"] in deps:
+        # Ensure we are installing the testbed-native version of
+        # build-essential by specifying :native. This is needed because the
+        # build-essential package is not Multi-Arch: foreign, so we need to
+        # make sure we are always picking up the package for the testbed
+        # architecture, not for the foreign test architecture.
+        deps.remove(['build-essential'])
+        deps.append(['build-essential:native'])
+        # Install the right crossbuild-essential package.
+        deps.append([f'crossbuild-essential-{test_arch}:native'])
+        # Add install libc-dev and libstdc++-dev, as a workaround for #815172.
+        deps.append([f'libc-dev:{test_arch}'])
+        deps.append([f'libstdc++-dev:{test_arch}'])
+
+    return (deps, package_under_test_deps)
 
 
 def _autodep8(srcdir):
@@ -546,6 +597,7 @@ def parse_debian_source(
     srcdir: str,
     testbed_caps: List[str],
     test_arch: str,
+    test_arch_is_foreign: bool,
     control_path: Optional[str] = None,
     auto_control: bool = True,
     ignore_restrictions: Optional[List[str]] = None,
@@ -635,11 +687,12 @@ def parse_debian_source(
                 test_names = record['Tests'].replace(',', ' ').split()
                 if len(test_names) == 0:
                     raise InvalidControl('*', '"Tests" field is empty')
-                (depends, synth_depends) = _expand_test_depends(
+                (depends, package_under_test_depends) = _expand_test_depends(
                     test_names[0],
                     record.get('Depends', '@'),
                     srcdir,
                     test_arch,
+                    test_arch_is_foreign,
                 )
                 if 'Test-command' in record:
                     raise InvalidControl('*', 'Only one of "Tests" or '
@@ -655,7 +708,8 @@ def parse_debian_source(
                         _check_architecture(n, test_arch, architectures)
 
                         test = Test(n, os.path.join(test_dir, n), None,
-                                    restrictions, features, depends, synth_depends)
+                                    restrictions, features, depends,
+                                    package_under_test_depends)
                         test.check_testbed_compat(testbed_caps, ignore_restrictions)
                     except Unsupported as u:
                         if (not only_tests) or n in only_tests:
@@ -665,11 +719,12 @@ def parse_debian_source(
                         tests.append(test)
             elif 'Test-command' in record:
                 command = record['Test-command']
-                (depends, synth_depends) = _expand_test_depends(
+                (depends, package_under_test_depends) = _expand_test_depends(
                     command,
                     record.get('Depends', '@'),
                     srcdir,
                     test_arch,
+                    test_arch_is_foreign,
                 )
                 if feature_test_name is None:
                     command_counter += 1
@@ -679,7 +734,7 @@ def parse_debian_source(
                 _debian_check_unknown_fields(name, record)
                 _check_architecture(name, test_arch, architectures)
                 test = Test(name, None, command, restrictions, features,
-                            depends, synth_depends)
+                            depends, package_under_test_depends)
                 test.check_testbed_compat(testbed_caps, ignore_restrictions)
                 tests.append(test)
             else:
